@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-from logging import error, exception
+from logging import getLogger
 from socket import (
     AF_INET,
     IPPROTO_TCP,
@@ -10,15 +10,20 @@ from socket import (
     TCP_KEEPCNT,
     TCP_KEEPIDLE,
     TCP_KEEPINTVL,
+    SHUT_RDWR,
     getaddrinfo,
     socket,
 )
 from threading import Timer
 from time import sleep
-from ph_wsgiref.simple_server import WSGIRequestHandler
+from signal import signal, SIGTERM
 
 from .resource_balancer import ResourceBalancer
 
+from ph_wsgiref.simple_server import WSGIRequestHandler
+
+LOGGER = getLogger('pothead.worker')
+
 
 class Handler(WSGIRequestHandler):
     protocol_version = "HTTP/1.1"
@@ -29,6 +34,14 @@ class Handler(WSGIRequestHandler):
         finally:
             self.close_connection = True
 
+    def parse_request(self):
+        res = super().parse_request()
+
+        # This socket shall no longer be fast-closed on SIGTERM
+        socket = self.request
+        self.server.sockets_waiting_for_request.remove(socket)
+
+        return res
 
 class Server:
     ssl_context = None
@@ -37,6 +50,7 @@ class Server:
     server_address = "localhost"
     passthrough_errors = False
     shutdown_signal = False
+    running = True
 
     def __init__(self, addr, app):
         (host, port) = addr
@@ -51,12 +65,14 @@ class Server:
 
         self.balancer = LoadBalancer(host, port)
         self.app = app
+        self.sockets_waiting_for_request = set()
 
     def run(self):
-        while True:
+        while self.running:
             with socket(
                 AF_INET, SOCK_STREAM
             ) as s, self.balancer.acquire_addr() as addr:
+                self.sockets_waiting_for_request.add(s)
                 try:
                     s.connect(addr)
 
@@ -65,20 +81,38 @@ class Server:
                     s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 2)
                     s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
                 except Exception as e:
-                    error("POTHEAD Connect To Broker: %s", e)
+                    LOGGER.error("Connect To Broker: %s", e)
                     sleep(2)
                     continue
 
                 try:
-                    Handler(s, addr, server)
+                    Handler(s, addr, self)
                 except:
-                    exception("POTHEAD WSGIRequestHandler")
+                    LOGGER.exception("In WSGI request handling")
                     sleep(1)
                     continue
 
     def get_app(self):
         return self.app
 
+    def shutdown(self):
+        LOGGER.info("shutting down workers...")
+        self.running = False
+        for s in list(self.sockets_waiting_for_request):
+            s.shutdown(SHUT_RDWR)
+            s.close()
+
+
+def install_term_handler(f):
+    previous = []
+    def on_term(_signal, _stack):
+        f()
+        for p in previous:
+            p()
+    p = signal(SIGTERM, on_term)
+    if p:
+        previous.append(p)
+
 
 class LoadBalancer:
     REFRESH_INTERVAL = 5
@@ -98,8 +132,10 @@ class LoadBalancer:
             hosts = (sockaddr for _f, _t, _p, _c, sockaddr in results)
             self._balancer.provision(hosts)
         except:
-            exception("Failed to refresh endpoints for %s:%d", self.host, self.port)
-        Timer(self.REFRESH_INTERVAL, self.refresh).start()
+            LOGGER.exception("Failed to refresh endpoints for %s:%d", self.host, self.port)
+        t = Timer(self.REFRESH_INTERVAL, self.refresh)
+        t.setDaemon(True)
+        t.start()
 
 
 if __name__ == "__main__":
@@ -107,6 +143,7 @@ if __name__ == "__main__":
     from importlib import import_module
     from threading import Thread
     from sys import path
+    from logging import basicConfig, INFO
 
     path.insert(0, ".")
 
@@ -137,11 +174,20 @@ if __name__ == "__main__":
         help="The WSGI request handler to handle requests",
     )
     args = parser.parse_args()
+    basicConfig(level=INFO)
+
+    LOGGER.info("Initializing server for app %s", args.app)
 
     server = Server(args.connect, args.app)
 
+    install_term_handler(server.shutdown)
+
+    LOGGER.info("Starting up %d workers connecting to %s", args.workers, args.connect)
+
     workers = [Thread(target=server.run) for _ in range(args.workers)]
     for worker in workers:
         worker.start()
     for worker in workers:
         worker.join()
+
+    LOGGER.info("shut down")
